/*
  universe.cpp

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/

#include "universe.h"

namespace DMX {

/**
 * @brief Universe::Universe
 * @param timeout_period
 */
Universe::Universe(int timeout_period)
  : null_start_data({0})
  , rx_timeout_period_(timeout_period)
  , status_(DMX_NULL)
{
}


/**
 * @brief Universe::~Universe
 */
Universe::~Universe()
{
}


/**
 * @brief Universe::slot
 * @param address
 * @return
 */
uint8_t Universe::slot(const uint16_t address)
{
  if (address == 0)
    return 0;
  if (address > null_start_data.size() - 1)
    return 0;

  null_start_mutex.lock();
  uint8_t val = null_start_data[address];
  null_start_mutex.unlock();

  return val;
}


/**
 * @brief Universe::rxRate Rate (in Hz) at which the universe is being recv'd.
 * @return Hz
 *
 *  Calculated as the rolling mean of (the number of frames in
 *  the last rx_timeout_peroid ms) / rx_timeout_period seconds.
 */
double Universe::rxRate()
{
  rx_timeout_();
  //         updates     per      second
  return rx_times_.size() / (rx_timeout_period_ / 1000.0);
}


/**
 * @brief milliseconds since the last update
 * @return
 */
long Universe::age()
{
  if (!last_updated_.time_since_epoch().count())
    return -1;        // universe has never been seen

  auto now = std::chrono::system_clock::now();
  auto elapsed = std::chrono::duration_cast<std::chrono::milliseconds>(now - last_updated_);

  return elapsed.count();
}


/**
 * @brief Universe::setData Accept new data from receiver
 * @param data varaibly sized data, beginning with a start code.
 *
 * The only setData method used to establish rx rate.
 */
void Universe::setData(const std::vector<uint8_t>& data)
{
  // no data?
  if (data.empty())
    return;

  // Alternate Start Code
  if (data.front() != E111_NULL_START)
    return altSCdata(data);

  // accept variable lenth input, but never more than the limit
  const auto max_length = null_start_data.size();
  const auto length = data.size() < max_length ? data.size() : max_length;
  null_start_mutex.lock();                                      // take the lock
  if (length < max_length)
    null_start_data.fill(0);                                    // wipe old data
  std::copy_n(data.cbegin(), length, null_start_data.begin());  // copy new data
  null_start_mutex.unlock();                                    // free the lock

  rx_timeout_(true);                                            // update rx times
  doDataCallbacks();                                            // notify callbacks
}


/**
 * @brief Universe::altSCdata
 * @param data
 *
 * \attention re-impliment this method to rx custom alternate start code data.
 */
void Universe::altSCdata(const std::vector<uint8_t> & data)
{
  switch (data.front())
  {
    case E111_ASC_TEXT_ASCII:
      break;
    case E111_ASC_TEST:
      break;
    case E111_ASC_TEXT_UTF8:
      break;
    case E111_ASC_MANUFACTURER:
      break;
    case E111_ASC_SIP:
      break;
  }
}


/**
 * @brief Universe::status
 * @return
 *
 * Like Schrödinger's cat, the status of the universe may exist in a superpostion
 * of states. Only by observing it will it collaps into a single status.
 */
uint8_t Universe::status()
{
  rx_timeout_(); // many things may have happened since the last observation
  return status_;
}


/**
 * @brief Universe::setStatus
 * @param val
 */
void Universe::setStatus(uint8_t val)
{
  if (val == RX_TIMEOUT)
    val = DMX_LOST;

  if (val == status_)
    return;

  status_ = val;
  doStatusCallbacks();
}


/**
 * @brief Universe::onStatusChange
 * @param cb
 * @return
 */
std::shared_ptr<void> Universe::onStatusChange(std::function<void()> cb)
{
  // wrap the callback with a shared pointer
  auto sp = std::make_shared<std::function<void()>>(std::move(cb));
  // add callback to list (as a weak pointer)
  cb_statusChange.push_back(sp);
  // return token that caller must keep throughout it's scope
  return sp;
}


/**
 * @brief Universe::doStatusCallbacks
 */
void Universe::doStatusCallbacks()
{
  for (auto it = cb_statusChange.begin(); it != cb_statusChange.end();)
    {
      if (auto sp = it->lock())
        { // if the caller is still holding the token
          (*sp)();
          ++it;
        }
      else
        { // or remove the callback
          it = cb_statusChange.erase(it);
        }
    }
}


/**
 * @brief Universe::setValue
 * @param address
 * @param value
 */
void Universe::setValue(const uint16_t address, const uint8_t value)
{
  if (address == 0)
    return;
  if (address > null_start_data.size() - 1)
    return;

  null_start_mutex.lock();
  null_start_data[address] = value;
  null_start_mutex.unlock();

  setStatus(DMX_ACTIVE);
  last_updated_ = std::chrono::system_clock::now();
  doDataCallbacks();
}


/**
 * @brief Universe::setValue
 * @param start
 * @param footprint
 * @param data
 */
void Universe::setValue(const uint16_t start, const uint16_t footprint,
                        const uint8_t* data)
{
  // start and footprint valid?
  if (start < 1 || start + footprint >= null_start_data.size())
    return;

  null_start_mutex.lock();
  std::copy_n(data, footprint, null_start_data.begin() + start);
  null_start_mutex.unlock();

  setStatus(DMX_ACTIVE);
  last_updated_ = std::chrono::system_clock::now();
  doDataCallbacks();
}


/**
 * @brief Universe::onData
 * @param cb
 * @return
 */
std::shared_ptr<void> Universe::onDataChange(const DataHandler cb)
{
  // wrap the callback with a shared pointer
  auto sp = std::make_shared<DataHandler>(std::move(cb));
  // add callback to list (as a weak pointer)
  cb_dataChange.push_back(sp);
  // return token that caller must keep throughout it's scope
  return sp;
}


/**
 * @brief Universe::doDataCallbacks
 */
void Universe::doDataCallbacks()
{
  for (auto it = cb_dataChange.begin(); it != cb_dataChange.end();)
  {
    if (auto sp = it->lock())
    { // if the caller is still holding the token
      (*sp)(this);
      ++it;
    }
    else
    { // or remove the callback
      it = cb_dataChange.erase(it);
    }
  }
}


/**
 * @brief Universe::rx_timeout_
 * @param add_now
 */
void Universe::rx_timeout_(bool add_now)
{
  auto now = std::chrono::system_clock::now();

  if (add_now)
    {
      setStatus(DMX_ACTIVE);
      last_updated_ = now;
      rx_times_.push(now);
    }
  else if (status_ == DMX_ACTIVE && age() >= E111_DATA_LOSS_TIMEOUT)
    // E1.11 DMX timeout period may be different the the rx_timeout_period
    setStatus(DMX_LOST);
  else if (age() >= rx_timeout_period_)
    // setStatus will be reimplimented to set the correct status
    setStatus(RX_TIMEOUT);

  auto elapsed = [] (auto& a, auto& b) {
      return std::chrono::duration_cast<std::chrono::milliseconds>(a - b).count();
    };

  long age;
  while (!rx_times_.empty())
  { // drop timestamps that are older than the timeout period
    age = elapsed(now, rx_times_.front());
    if (age < rx_timeout_period_)
      break;
    else
      rx_times_.pop();
  }
}

} // namespace DMX
