/*
  universe.h

  Copyright (c) 2020 Kevin Matz (kevin.matz@gmail.com)

  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files (the "Software"), to deal
  in the Software without restriction, including without limitation the rights
  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
  copies of the Software, and to permit persons to whom the Software is
  furnished to do so, subject to the following conditions:

  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  SOFTWARE.
*/
#pragma once

#include "dmx.h"

#include <chrono>
#include <cstdint>
#include <functional>
#include <memory>
#include <mutex>
#include <queue>
#include <vector>

namespace DMX {

class Universe; // forward declare the Univserse class
using DimmerData  = std::array<uint8_t, E111_LAST_SLOT + 1>; //!< Array of 513 bytes for basic data.
using DataHandler = std::function<void(Universe *)>;         //!< Callback function for data handling.

/**
 * @brief The Universe class
 *
 *  \cite DMX ANSI E1.11 describes may things about moving data over serial EIA-485-A
 *  links.  This class encapselates that data, regardless of how it was
 *  transmitted.
 */
class Universe {
  public:
    Universe (int timeout_period = E111_DATA_LOSS_TIMEOUT);
    virtual ~Universe ();

    virtual uint8_t slot (const uint16_t);
    virtual double rxRate();
    virtual long age();

    virtual void setValue (const uint16_t address, const uint8_t value);
    virtual void setValue (const uint16_t start, const uint16_t footprint,
                           const uint8_t* data);

    void setData (const std::vector<uint8_t> &);
    std::shared_ptr<void> onDataChange (const DataHandler);

    virtual void altSCdata(const std::vector<uint8_t> &);

    /**
     * @brief The Status enum
     */
    enum Status : uint8_t {
        DMX_NULL   = 0,  //!< uninitialized
        DMX_ACTIVE = 1,  //!< actively sending/receiving data
        DMX_LOST   = 2,  //!< no activity in E111_DATA_LOSS_TIMEOUT
        RX_TIMEOUT = 250 //!< no activity in the constructed timeout period
    };
    virtual void    setStatus(uint8_t);
    virtual uint8_t status();
    virtual std::shared_ptr<void> onStatusChange(std::function<void()>);

  protected:
    DimmerData null_start_data;           //!< NULL Start Code data
    mutable std::mutex null_start_mutex;  //!< memory protect Null Start data

    void doDataCallbacks();               //!< execute data changed callbacks
    void doStatusCallbacks();             //!< execute status changed callbacks

    std::chrono::system_clock::time_point last_updated_; //!< time of the latest update

  private:
    std::queue<std::chrono::time_point<std::chrono::system_clock>> rx_times_;

    void rx_timeout_(bool add_now = false);
    const int rx_timeout_period_;

    uint8_t status_;                      //!< the operating state of the universe
    std::vector<std::weak_ptr<std::function<void()>>> cb_statusChange;
    std::vector<std::weak_ptr<DataHandler>> cb_dataChange;
};

} // namespace DMX
